! WHIZARD 3.1.6 Mar 20 2025
!
! Copyright (C) 1999-2025 by
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!
!     with contributions from
!     cf. main AUTHORS file
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

submodule (electron_pdfs) electron_pdfs_s

  use iso_varying_string, string_t => varying_string
  use io_units
  use sm_qed, only: alpha_qed_t, alpha_qed_fixed_t, alpha_qed_from_scale_t
  use diagnostics
  use format_defs, only: FMT_12
  use numeric_utils, only: log_prec
  use physics_defs, only: EPDF_LL, EPDF_NLL
  use sm_physics, only: polylog, Li2, psimr, psir

  implicit none

contains

  module pure function coeffqed_b0(nf, nlep)
    integer, intent(in) :: nf, nlep
    real(wp) :: n_lep, coeffqed_b0
    n_lep = real(nlep, kind=wp)
    coeffqed_b0 = sumQ2q(nf) + n_lep/3._wp
    contains
    pure function sumQ2q (nf)
      integer, intent(in) :: nf
      real(wp) :: sumQ2q
      select case (nf)
      case (0)
         sumQ2q = 0.0_wp
      case (1)
         sumQ2q = 1.0_wp/9.0_wp
      case (2)
         sumQ2q = 5.0_wp/9.0_wp
      case (3)
         sumQ2q = 2.0_wp/3.0_wp
      case (4)
         sumQ2q = 10.0_wp/9.0_wp
      case (5)
         sumQ2q = 11.0_wp/9.0_wp
      case (6:)
         sumQ2q = 5.0_wp/3.0_wp
      end select
    end function sumQ2q
  end function coeffqed_b0

  module pure function coeffqed_b1(nf, nlep)
    integer, intent(in) :: nf, nlep
    real(wp) :: n_lep, coeffqed_b1
    n_lep = real(nlep, kind=wp)
    coeffqed_b1 = sumQ4q(nf) + n_lep/4._wp
    contains
    pure function sumQ4q (nf)
      integer, intent(in) :: nf
      real(wp) :: sumQ4q
      select case (nf)
      case (0)
         sumQ4q = 0.0_wp
      case (1)
         sumQ4q = 1.0_wp/81.0_wp
      case (2)
         sumQ4q = 17.0_wp/81.0_wp
      case (3)
         sumQ4q = 2.0_wp/9.0_wp
      case (4)
         sumQ4q = 34.0_wp/81.0_wp
      case (5)
         sumQ4q = 35.0_wp/81.0_wp
      case (6:)
         sumQ4q = 17.0_wp/27.0_wp
      end select
    end function sumQ4q
  end function coeffqed_b1

  module subroutine qed_pdf_init &
       (qed_pdf, mass, alpha, charge, q_max, order, log_order, nlep)
    class(qed_pdf_t), intent(out) :: qed_pdf
    real(default), intent(in) :: mass, alpha, q_max, charge
    integer, intent(in) :: order, log_order, nlep
    qed_pdf%mass = real(mass,kind=wp)
    qed_pdf%q_max = real(q_max,kind=wp)
    qed_pdf%alpha = real(alpha,kind=wp)
    qed_pdf%order = order
    qed_pdf%log_order = log_order
    qed_pdf%nlep = nlep
    qed_pdf%running = .false.
    qed_pdf%eps = alpha/pi * real(charge,kind=wp)**2 &
         * (2._wp * log (q_max / mass) - 1._wp)
  end subroutine qed_pdf_init

  module subroutine qed_pdf_write (qed_pdf, unit, with_qed)
    class(qed_pdf_t), intent(in) :: qed_pdf
    integer, intent(in), optional :: unit
    logical, intent(in), optional :: with_qed
    integer :: u
    logical :: show_qed
    u = given_output_unit (unit)
    show_qed = .false.
    if (present (with_qed))  show_qed = with_qed
    write (u, "(3x,A)")  "QED structure function (PDF):"
    write (u, "(5x,A,I0)") "Flavor    = ", qed_pdf%flv
    write (u, "(5x,A," // FMT_12 // ")") "Mass      = ", qed_pdf%mass
    write (u, "(5x,A," // FMT_12 // ")") "q_max     = ", qed_pdf%q_max
    write (u, "(5x,A," // FMT_12 // ")") "alpha     = ", qed_pdf%alpha
    write (u, "(5x,A,I0)") "Order     = ", qed_pdf%order
    write (u, "(5x,A,I0)") "Log. ord. = ", qed_pdf%log_order
    write (u, "(5x,A,I0)") "# leptons = ", qed_pdf%nlep
    write (u, "(5x,A,I0)") "Run. Coupling = ", qed_pdf%running
    write (u, "(5x,A," // FMT_12 // ")") "epsilon   = ", qed_pdf%eps
    if (show_qed) then
       call qed_pdf%aqed%write (u)
    end if
  end subroutine qed_pdf_write

  module subroutine qed_pdf_set_order (qed_pdf, order)
    class(qed_pdf_t), intent(inout) :: qed_pdf
    integer, intent(in) :: order
    qed_pdf%order = order
  end subroutine qed_pdf_set_order

  module subroutine qed_pdf_evolve_qed_pdf (qed_pdf, x, xb, rb, ff)
    class(qed_pdf_t), intent(inout) :: qed_pdf
    real(default), intent(in) :: x, xb, rb
    real(default), intent(inout) :: ff
    real(default), parameter :: &
         & xmin = 0.00714053329734592839549879772019_default
    real(default), parameter :: &
         g1 = 3._default / 4._default, &
         g2 = (27 - 8 * pi2) / 96._default, &
         g3 = (27 - 24 * pi2 + 128 * zeta3) / 384._default
    real(default) :: x_2, log_x, log_xb
    if (ff > 0 .and. qed_pdf%order > 0) then
       ff = ff * (1 + g1 * qed_pdf%eps)
       x_2 = x * x
       if (rb > 0)  ff = ff * (1 - (1-x_2) / (2 * rb))
       if (qed_pdf%order > 1) then
          ff = ff * (1 + g2 * qed_pdf%eps**2)
          if (rb > 0 .and. xb > 0 .and. x > xmin) then
             log_x  = log_prec (x, xb)
             log_xb = log_prec (xb, x)
             ff = ff * (1 - ((1 + 3 * x_2) * log_x + xb * (4 * (1 + x) * &
                  log_xb + 5 + x)) / (8 * rb) * qed_pdf%eps)
          end if
          if (qed_pdf%order > 2) then
             ff = ff * (1 + g3 * qed_pdf%eps**3)
             if (rb > 0 .and. xb > 0 .and. x > xmin) then
                ff = ff * (1 - ((1 + x) * xb &
                     * (6 * Li2(x) + 12 * log_xb**2 - 3 * pi2) &
                     + 1.5_default * (1 + 8 * x + 3 * x_2) * log_x &
                     + 6 * (x + 5) * xb * log_xb &
                     + 12 * (1 + x_2) * log_x * log_xb &
                     - (1 + 7 * x_2) * log_x**2 / 2 &
                     + (39 - 24 * x - 15 * x_2) / 4) &
                     / (48 * rb) * qed_pdf%eps**2)
             end if
          end if
       end if
    end if
  end subroutine qed_pdf_evolve_qed_pdf

  module subroutine set_qed_pdf_parameters &
      (epdf, scale, alpha, running, nlep, nf, ln0, eta0, p, al0_2pi, al_2pi, n, run, order)
    type(qed_pdf_t), intent(in) :: epdf
    real(wp), intent(in) :: scale
    real(wp), intent(out) :: ln0, eta0, p, alpha, al0_2pi, al_2pi, n, run
    logical, intent(out) :: running
    integer, intent(out) :: nlep, nf
    logical, dimension(6), intent(out) :: order
    order = .false.
    running=epdf%running
    alpha=epdf%alpha
    al0_2pi = alpha / 2._wp / pi
    if (allocated (epdf%q_in)) then
       ln0 = 2._wp*log(epdf%q_in/epdf%mass)
       eta0 = alpha/pi * 2._wp*log(scale/epdf%q_in)
    else
       ln0 = 0._wp
       eta0 = alpha/pi * 2._wp*log(scale/epdf%mass)
    end if
    if (running) then
       p = t_alpha (epdf, scale)
       run = 1._wp
    else
       p = eta0 / 2._wp
       run = 0._wp
    end if
    select case (epdf%log_order)
    case (EPDF_LL)
       order(1:4) = .true.
       nlep = epdf%nlep
       nf = 0
       al_2pi = 0._wp
    case (EPDF_NLL)
       order(1:6) = .true.
       if (running) then
          select type (aqed => epdf%aqed)
          type is (alpha_qed_from_scale_t)
             nf = aqed%nf
             nlep = aqed%nlep
             al_2pi = aqed%get(scale) / 2._wp / pi
          type is (alpha_qed_fixed_t)
             call msg_fatal &
                  ("Object has to be called with running alpha.")
          end select
       else
          nlep = epdf%nlep
          al_2pi = al0_2pi
       end if
    end select
    n = real(nlep, kind=wp)
  end subroutine set_qed_pdf_parameters

  module pure function elec_asym &
      (xb, nlep, nf, n, p, al0_2pi, al_2pi, order, running) &
      result (elec_as)
    real(wp), dimension(7,7,2) :: elec_as
    real(wp), intent(in) :: xb
    real(wp), intent(in) :: n, p, al0_2pi, al_2pi
    integer, intent(in) :: nf, nlep
    logical, intent(in) :: running
    logical, intent(in), dimension(6) :: order
    real(wp), parameter :: lambda0 = 3._wp/4._wp
    real(wp) :: lambda1, xi0, xi1, xihat1, fac, ex0
    real(wp) :: ca, cb, b0, b01
    elec_as=0.0_wp
    xi0 = 2._wp*p
    if ( order(5) .eqv. .false. ) then
       elec_as(7,4,1) = exp((lambda0 - eulerc)*xi0) / gamma (xi0) * &
            xb**(xi0 - 1._wp)
    else
       lambda1 = 3._wp/16._wp - pi2/4._wp + 3._wp*zeta3 - &
            n*(3._wp + 4._wp*pi2)/36._wp
       if (running) then
          b0  = coeffqed_b0(nf,nlep)
          b01 = coeffqed_b1(nf,nlep)/b0
          ex0 = al_2pi*(1._wp - exp(-b0*xi0))/b0
          xi1 = xi0 - 2._wp*ex0*(5._wp/9._wp*n + b01)
          xihat1 = xi0*lambda0 + ex0*(lambda1 - 3._wp/2._wp*b01)
       else
          xi1 = xi0*(1._wp - 10._wp/9._wp*al_2pi*n)
          xihat1 = xi0*(lambda0 + al_2pi*lambda1)
       end if
       fac = exp(xihat1 - eulerc*xi1) * xb**(xi1 - 1._wp) / gamma(xi1)
       ca = eulerc + psir(xi1) - 0.5_wp
       cb = 2._wp*(ca**2 - psimr(xi1,1)) + pi2/3._wp - 2.5_wp
       elec_as(7,4,1) = fac*(1._wp - al0_2pi*cb)
       elec_as(7,4,2) = fac*(0.25_wp - ca)
       elec_as(7,5,2) = fac*2._wp*al0_2pi
       elec_as(7,5,1) = 2._wp*ca*elec_as(7,5,2)
       elec_as(7,6,1) = -elec_as(7,5,2)
    end if
  end function elec_asym

  module pure function phot_asym &
       (x, xb, log_xb, n, p, al0_2pi, order, running) result (phot_as)
    real(wp), dimension(7,7,2) :: phot_as
    real(wp), intent(in) :: x, xb, log_xb
    real(wp), intent(in) :: p, n, al0_2pi
    logical, intent(in) :: running
    logical, intent(in),  dimension(6) :: order
    real(wp) :: lambda1, xi10, xihat10
    real(wp) :: xi0, den
    real(wp) :: mf1k, mf10, ca, cb, fac0, fac
    real(wp) :: d1, d2, cc, k0
    phot_as = 0.0_wp
    xi0 = 2._wp*p
    if (order(5) .eqv. .false. ) then
       den = -2._wp/3._wp*n - 3._wp/2._wp - 2._wp*log_xb
       mf10 = 1._wp - 2._wp/3._wp*pi2/den**2
       mf1k = 1._wp - xi0*(pi2/3._wp - 2._wp*zeta3*xi0)/den - &
         (2._wp/3._wp*pi2 - (8._wp*zeta3 - pi4/45._wp*xi0)*xi0)/den**2
       phot_as(7,4,1) = ( &
           exp(-xi0*n/3._wp)*mf10 - &
           exp((3._wp/4._wp - eulerc + log_xb)*xi0)/gamma(1._wp+xi0)*mf1k &
         )/den
    else
       lambda1 = 1._wp/4._wp - pi2/3._wp + 4._wp*zeta3 - &
            n*(3._wp + 4._wp*pi2)/27._wp
       ca = 5._wp + 4._wp/3._wp*n
       if (running) then
         xi10 = 2._wp*(1._wp - al0_2pi*(10._wp/9._wp*n + 3._wp/2._wp))
         xihat10 = 3._wp/2._wp*(1._wp + al0_2pi*(lambda1 - 3._wp/2._wp))
         cb = (pi2/2._wp - 22._wp + 8._wp*ca)/3._wp + 3._wp/2._wp
         fac0 = 2._wp*exp(-n*xi0/3._wp)
         d2 = -((2._wp/3._wp)*n + xihat10)/xi10
         d1 = d2 - (2._wp/3._wp)*n/xi10
         cc = exp(-n*xi0/3._wp)
       else
         xi10 = 2._wp*(1._wp - 10._wp/9._wp*al0_2pi*n)
         xihat10 = 3._wp/2._wp*(1._wp + al0_2pi*lambda1)
         cb = (pi2/2._wp - 22._wp + 8._wp*ca)/3._wp
         fac0 = 2._wp*exp(-(2._wp/3._wp + al0_2pi)*n*p)
         d2 = -((2._wp/3._wp + al0_2pi)*n + xihat10)/xi10
         d1 = d2
         cc = 1._wp
       end if
       k0 = xi10*p
       phot_as(7,:,:) = sum_rm(log_xb, al0_2pi, ca, cb, d1, d2, cc, k0)/xi10
       phot_as(7,3,1) = -al0_2pi*(2._wp*xb/x + x)
       phot_as(7,4,1) = phot_as(7,4,1) + phot_as(7,3,1)/2._wp
       phot_as(7,4,2) = phot_as(7,4,2) - phot_as(7,4,1)
       phot_as = fac0*phot_as
    end if
  end function phot_asym

  module function recasym &
       (flv, x, xb, log_xb, nlep, nf, n, p, &
       al0_2pi, al_2pi, order, running) result (jll_nll)
    integer, intent(in) :: flv, nlep, nf
    real(wp), intent(in) :: x, xb, log_xb
    real(wp), intent(in) :: n, p, al0_2pi, al_2pi
    logical, intent(in) :: running
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2) :: jll_nll
    jll_nll = 0.0
    select case (flv)
    case (EPDF_ELE,EPDF_S,EPDF_NS)
      jll_nll = elec_asym &
         (xb, nlep, nf, n, p, al0_2pi, al_2pi, order, running)
    case (EPDF_POS)
      jll_nll = 0._wp
    case (EPDF_G)
      jll_nll = phot_asym &
         (x, xb, log_xb, n, p, al0_2pi, order, running)
    case default
       call msg_fatal &
            ("recasym: wrong lepton flavor.")
    end select
  end function recasym

  module function bar_asym (flv, xb, n, run, order) result (jll_nll)
    integer, intent(in) :: flv
    real(wp), intent(in) :: xb
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2) :: jll_nll
    select case (flv)
    case (EPDF_ELE,EPDF_S,EPDF_NS)
       call elecbar_asym_p (xb, jll_nll, n, run, order)
    case (EPDF_POS)
       jll_nll = 0._wp
    case (EPDF_G)
       call photbar_asym_p (jll_nll, n, run, order)
    case default
       call msg_fatal &
            ("bar_asym: wrong lepton flavor.")
    end select
  end function bar_asym

  module function recnum &
       (flv, x, xb, xp, pl, n) result (jll_nll)
    integer, intent(in) :: flv
    real(wp), dimension(7,7,2) :: jll_nll
    real(wp), intent(in) :: x, xb, xp, n
    real(wp), dimension(12), intent(in) :: pl
    jll_nll = 0._wp
    select case (flv)
    case (EPDF_S)
     jll_nll(6,:,1) = endpoint_func_S (x, xb, xp, pl, n)
    case (EPDF_NS)
     jll_nll(6,:,1) = endpoint_func_NS (x, xb, xp, pl, n)
    case (EPDF_G)
     jll_nll(6,:,1) = endpoint_func_GAM (x, xb, xp, pl)
    case (EPDF_ELE)
     jll_nll(6,:,1) = endpoint_func_ELE (x, xb, xp, pl, n)
    case (EPDF_POS)
     jll_nll(6,:,1) = endpoint_func_POS (x, xb, xp, pl, n)
    case default
       call msg_fatal &
            ("recnum: wrong lepton flavor.")
    end select
  end function recnum

  module function recbar &
       (flv, x, xb, n, run, order) result (jll_nll)
    integer, intent(in) :: flv
    real(wp), intent(in) :: x, xb
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2) :: jll_nll
    select case (flv)
    case (EPDF_S)
       call recbar_singlet (x, xb, jll_nll, n, run, order)
    case (EPDF_NS)
       call recbar_nonsinglet (x, xb, jll_nll, n, run, order)
    case (EPDF_G)
       call recbar_photon (jll_nll, x, xb, n, run, order)
    case (EPDF_ELE)
       call recbar_ele (x, xb, jll_nll, n, run, order)
    case (EPDF_POS)
       call recbar_pos (jll_nll, n, run, order)
    case default
       call msg_fatal &
            ("recbar: wrong lepton flavor.")
    end select
  end function recbar

  module function rechat &
       (flv, x, xb, xp,  pl, n, run, order) result (jll_nll)
    integer, intent(in) :: flv
    real(wp), intent(in) :: x, xb, xp
    real(wp), dimension(12), intent(in) :: pl
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2) :: jll_nll
    select case (flv)
    case (EPDF_S)
       call rechat_singlet (x, xb, xp, pl, jll_nll, n, run, order)
    case (EPDF_G)
       call rechat_photon (x, xb, xp, pl, jll_nll, n, run, order)
    case (EPDF_NS)
       call rechat_nonsinglet (x, xb, xp, pl, jll_nll, n, run, order)
    case (EPDF_ELE)
       call rechat_ele (x, xb, xp, pl, jll_nll, n, run, order)
    case (EPDF_POS)
       call rechat_pos (x, xb, xp, pl, jll_nll, n, run, order)
    case default
       call msg_fatal &
            ("rechat: wrong lepton flavor.")
    end select
  end function rechat

  module subroutine elecbar_asym_p (xb, jll_nll, n, run, order)
    real(wp), intent(in) :: xb
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    jll_nll = 0._wp
    if (order(1)) then
      jll_nll(1,4,1)=2._wp
    end if
    if (order(2)) then
      jll_nll(2,4,1)=6._wp
      jll_nll(2,5,1)=8._wp
    end if
    if (order(3)) then
      jll_nll(3,4,1)=27._wp/2._wp - 4._wp*pi2
      jll_nll(3,5,1)=36._wp
      jll_nll(3,6,1)=24._wp
    end if
    if (order(4)) then
      jll_nll(4,4,1)=-2._wp
      jll_nll(4,4,2)=2._wp
      jll_nll(4,5,1)=-4._wp
    end if
    if (order(5)) then
      jll_nll(5,4,1)=1._wp- &
        4._wp*(5._wp/9._wp*n - pi2/3._wp - (n/3._wp-3._wp/4._wp)*run)
      jll_nll(5,4,2)=2._wp*(3._wp-2._wp/3._wp*n*run)
      jll_nll(5,5,1)=-2._wp*(7._wp-4._wp/3._wp*n*run)
      jll_nll(5,5,2)=8._wp
      jll_nll(5,6,1)=-12._wp
    end if
    if (order(6)) then
      jll_nll(6,4,1)=2._wp*(4.5_wp - 20._wp*zeta3 + 3._wp*pi2 - &
      n*(11._wp+ (4._wp/3._wp)*pi2)/3._wp - &
      2._wp*(4._wp + 4._wp/9._wp*pi2 + 2._wp/9._wp*n**2 - 1._wp/27._wp*n)*run)
      jll_nll(6,4,2)= 2._wp*(6.75_wp - 2._wp*pi2 - &
         4._wp*(1._wp - n/3._wp)*n*run)
      jll_nll(6,5,1)=-2._wp*(8.5_wp + (80._wp/9._wp)*n - (20._wp/3._wp)*pi2 + &
        4._wp*(3._wp - 7._wp/3._wp*n + 2._wp/9._wp*n**2)*run)
      jll_nll(6,5,2)=4._wp*(9._wp - 8._wp/3._wp*n*run)
      jll_nll(6,6,1)=-12._wp*(5._wp - 4._wp/3._wp*n*run)
      jll_nll(6,6,2)=24._wp
      jll_nll(6,7,1)=-32._wp
    end if
    jll_nll = jll_nll/xb
  end subroutine elecbar_asym_p

  module subroutine photbar_asym_p (jll_nll, n, run, order)
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    jll_nll = 0._wp
    if (order(1)) then
      jll_nll(1,4,1)=1._wp
    end if
    if (order(2)) then
      jll_nll(2,4,1)=1.5_wp - (2._wp * n)/3._wp
      jll_nll(2,5,1)=2._wp
    end if
    if (order(3)) then
      jll_nll(3,4,1)=2.25_wp - n + 4._wp * n**2 / 9._wp - (2._wp*pi2)/3._wp
      jll_nll(3,5,1)=6._wp - (4._wp*n)/3._wp
      jll_nll(3,6,1)=4._wp
    end if
    if (order(4)) then
      jll_nll(4,4,1)=- 1._wp
      jll_nll(4,4,2)=1._wp
    end if
    if (order(5)) then
      jll_nll(5,4,1)=-4._wp - 26._wp/9._wp*n + 2._wp*(n/3._wp - 3._wp/4._wp)*run
      jll_nll(5,4,2)=3._wp/2._wp - 2._wp/3._wp*n - 2._wp/3._wp*n*run
      jll_nll(5,5,1)=-7._wp - 4._wp/3._wp*n
      jll_nll(5,5,2)=2._wp
      jll_nll(5,6,1)=-3._wp
    end if
    if (order(6)) then
      jll_nll(6,4,1)= -5.625_wp + 11._wp/6._wp*pi2 - 6._wp*zeta3 + &
         (2._wp/9._wp*pi2 - 23._wp/6._wp)*n + 52._wp * n**2/27._wp + &
         (13._wp/3._wp*n - 9._wp/2._wp + 28._wp/27._wp*n**2)*run
      jll_nll(6,4,2)=2.25_wp - 2._wp/3._wp*pi2 - &
         n + 4._wp/9._wp*n**2 + 2._wp*(8._wp/9._wp*n - 1._wp)*n*run
      jll_nll(6,5,1)=-18.5_wp + 2._wp*pi2 + 8._wp/9._wp*n**2 - 20._wp/3._wp*n + &
         2._wp*(3._wp*n - 3._wp + 4._wp/9._wp*n**2)*run
      jll_nll(6,5,2)=2._wp*(3._wp - 2._wp*n/3._wp - 4._wp/3._wp*n*run)
      jll_nll(6,6,1)=-18.5_wp - 2._wp/3._wp*n + 10._wp/3._wp*n*run
      jll_nll(6,6,2)=4._wp
      jll_nll(6,7,1)=-6._wp
    end if
  end subroutine photbar_asym_p

  module subroutine recbar_singlet (x, xb, jll_nll, n, run, order)
    real(wp), intent(in) :: x, xb
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    real(wp), dimension(7,7,2) :: jll_nll_pos, jll_nll_ele
    call recbar_ele (x, xb, jll_nll_ele, n, run, order)
    call recbar_pos (jll_nll_pos, n, run, order)
    jll_nll = add_logvec(jll_nll_ele,jll_nll_pos)
  end subroutine recbar_singlet

  module subroutine recbar_nonsinglet (x, xb, jll_nll, n, run, order)
    real(wp), intent(in) :: x, xb
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    real(wp), dimension(7,7,2) :: jll_nll_pos, jll_nll_ele
    call recbar_ele (x, xb, jll_nll_ele, n, run, order)
    call recbar_pos (jll_nll_pos, n, run, order)
    jll_nll = sub_logvec(jll_nll_ele,jll_nll_pos)
  end subroutine recbar_nonsinglet

  module subroutine recbar_photon (jll_nll, x, xb, n, run, order)
    real(wp), intent(in) :: x, xb
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    jll_nll = 0._wp
    if (x .le. 0.1e-5_wp) then
      if (order(1)) then
        jll_nll(1,4,1) = 1._wp
      end if
      if (order(2)) then
        jll_nll(2,4,1) = (9._wp - 4._wp*n)/6._wp
      end if
      if (order(3)) then
        jll_nll(3,4,1) = (-4._wp*(9._wp - 4._wp*n)*n + 3._wp*(27._wp - 8._wp*pi2))/36._wp
        jll_nll(3,4,1) = (-4._wp*(9._wp - 4._wp*n)*n + 3._wp*(27._wp - 8._wp*pi2))/36._wp
        jll_nll(3,5,1) = (2._wp*(9._wp - 2._wp*n))/3._wp
        jll_nll(3,6,1) = 4._wp
      end if
      if (order(4)) then
        jll_nll(4,4,1) = -1._wp
        jll_nll(4,4,1) = -1._wp
      end if
      if (order(5)) then
        jll_nll(5,4,1) = (-4._wp*n*(13._wp - 3._wp*run) - 9._wp*(8._wp + 3._wp*run))/18._wp
        jll_nll(5,4,2) = (9._wp - 4._wp*n*(1._wp + run))/6._wp
      end if
      if (order(6)) then
        jll_nll(6,4,1)=(4._wp*n*(8._wp*n*(13._wp + 7._wp*run) - &
          3._wp*(69._wp - 2._wp*(2._wp*pi2 + 39._wp*run))) - &
          9._wp*(135._wp - 4._wp*(11._wp*pi2 + 9._wp*(-3._wp*run - 4._wp*zeta3))))/216._wp
        jll_nll(6,4,2) = (3._wp*(27._wp - 8._wp*pi2) + &
          4._wp*n*(-9._wp*(1._wp + 2._wp*run) + 4._wp*n*(1._wp + 3._wp*run)))/36._wp
      end if
    elseif (xb .le. 1.0e-7_wp) then
      if (order(1)) then
        jll_nll(1,4,1) = 1._wp
      end if
      if (order(2)) then
        jll_nll(2,4,1) = (9._wp - 4._wp*n)/6._wp
        jll_nll(2,5,1) = 2._wp
      end if
      if (order(3)) then
        jll_nll(3,4,1) = (-4._wp*(9._wp - 4._wp*n)*n + 3._wp*(27._wp - 8._wp*pi2))/36._wp
        jll_nll(3,5,1) = (2._wp*(9._wp - 2._wp*n))/3._wp
        jll_nll(3,6,1) = 4._wp
      end if
      if (order(4)) then
        jll_nll(4,4,1) = -1._wp
      end if
      if (order(5)) then
        jll_nll(5,4,1) = (-4._wp*n*(13._wp - 3._wp*run) - 9._wp*(8._wp + 3._wp*run))/18._wp
        jll_nll(5,5,1) = (-21._wp - 4._wp*n)/3._wp
        jll_nll(5,6,1) = -3._wp
        jll_nll(5,4,2) = (9._wp - 4._wp*n*(1._wp + run))/6._wp
      end if
      if (order(6)) then
        jll_nll(6,4,1)=(4._wp*n*(8._wp*n*(13._wp + 7._wp*run) - &
          3._wp*(69._wp - 2._wp*(2._wp*pi2 + 39._wp*run))) - 9._wp*(135._wp - &
          4._wp*(11._wp*pi2 + 9._wp*(-3._wp*run - 4._wp*zeta3))))/216._wp
        jll_nll(6,5,1)=(-9._wp*(37._wp - 4._wp*(pi2 - 3._wp*run)) + &
          4._wp*n*(-3._wp*(10._wp - 9._wp*run) + 4._wp*n*(1._wp + run)))/18._wp
        jll_nll(6,6,1) = (-111._wp - 4._wp*n*(1._wp - 5._wp*run))/6._wp
        jll_nll(6,7,1) = -6._wp
        jll_nll(6,4,2) = (3._wp*(27._wp - 8._wp*pi2) + &
          4._wp*n*(-9._wp*(1._wp + 2._wp*run) + 4._wp*n*(1._wp + 3._wp*run)))/36._wp
        jll_nll(6,5,2) = (2._wp*(9._wp - 2._wp*n*(1._wp + 2._wp*run)))/3._wp
      end if
    else
      if (order(1)) then
        jll_nll(1,4,1)=1._wp
      end if
      if (order(2)) then
        jll_nll(2,4,1)=3._wp/2._wp - (2._wp*n)/3._wp
        jll_nll(2,5,1)=2._wp
      end if
      if (order(3)) then
        jll_nll(3,4,1)=9._wp/4._wp + (-1._wp + (4._wp*n)/9._wp)*n - (2._wp*pi2)/3._wp
        jll_nll(3,5,1)=6._wp - (4._wp*n)/3._wp
        jll_nll(3,6,1)=4._wp
      end if
      if (order(4)) then
        jll_nll(4,4,1)=-1._wp
        jll_nll(4,4,2)=1._wp
      end if
      if (order(5)) then
        jll_nll(5,4,1) = (-4._wp*n*(13._wp - 3._wp*run) - 9._wp*(8._wp + 3._wp*run))/18._wp
        jll_nll(5,5,1) = (-21._wp - 4._wp*n)/3._wp
        jll_nll(5,6,1) = -3._wp
        jll_nll(5,4,2) = (9._wp - 4._wp*n*(1._wp + run))/6._wp
      end if
      if (order(6)) then
        jll_nll(6,4,1)=(4._wp*n*(8._wp*n*(13._wp + 7._wp*run) - 3._wp*(69._wp - &
          2._wp*(2._wp*pi2 + 39._wp*run))) - 9._wp*(135._wp - 4._wp*(11._wp*pi2 + &
          9._wp*(-3._wp*run - 4._wp*zeta3))))/216._wp
        jll_nll(6,5,1)=(-9._wp*(37._wp - 4._wp*(pi2 - 3._wp*run)) + &
          4._wp*n*(-3._wp*(10._wp - 9._wp*run) + 4._wp*n*(1._wp + run)))/18._wp
        jll_nll(6,6,1) = (-111._wp - 4._wp*n*(1._wp - 5._wp*run))/6._wp
        jll_nll(6,7,1) = -6._wp
        jll_nll(6,4,2) = (3._wp*(27._wp - 8._wp*pi2) + 4._wp*n*(-9._wp*(1._wp + &
          2._wp*run) + 4._wp*n*(1._wp + 3._wp*run)))/36._wp
        jll_nll(6,5,2) = (2._wp*(9._wp - 2._wp*n*(1._wp + 2._wp*run)))/3._wp
        jll_nll(6,6,2) = 4._wp
      end if
    end if
  end subroutine recbar_photon

  module subroutine recbar_ele (x, xb, jll_nll, n, run, order)
    real(wp), intent(in) :: x, xb
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    jll_nll = 0._wp
    if (x .le. 0.1e-5_wp) then
      if (order(1)) then
        jll_nll(1,4,1) = 2._wp*x
      end if
      if (order(2)) then
        jll_nll(2,4,1) = 4._wp
      end if
      if (order(3)) then
        jll_nll(3,4,1) = 18._wp
      end if
      if (order(4)) then
        jll_nll(4,4,1) = -2._wp*x
        jll_nll(4,4,2) = 2._wp*x
      end if
      if (order(5)) then
        jll_nll(5,4,1) = (-3._wp + 4._wp*n)/3._wp
        jll_nll(5,4,2) = 4._wp
      end if
      if (order(6)) then
        jll_nll(6,4,1) = (3._wp*(15._wp + 4._wp*(2._wp*pi2 - 9._wp*run)) - &
          2._wp*n*(22._wp - (15._wp - 4._wp*n)*run))/9._wp
        jll_nll(6,4,2) = (2._wp*(27._wp - 8._wp*n*run))/3._wp
      end if
    elseif (xb .le. 1.0e-7_wp) then
      if (order(1)) then
        jll_nll(1,4,1) = 2._wp/xb
      end if
      if (order(2)) then
        jll_nll(2,4,1) = 6._wp/xb
        jll_nll(2,5,1) = 8._wp/xb
      end if
      if (order(3)) then
        jll_nll(3,4,1) = (27._wp - 8._wp*pi2)/(2._wp*xb)
        jll_nll(3,5,1) = 36._wp/xb
        jll_nll(3,6,1) = 24._wp/xb
      end if
      if (order(4)) then
        jll_nll(4,4,1) = -2._wp/xb
        jll_nll(4,5,1) = -4._wp/xb
        jll_nll(4,4,2) = 2._wp/xb
      end if
      if (order(5)) then
        jll_nll(5,4,1) = (3._wp*(4._wp*pi2 + 3._wp*(1._wp - 3._wp*run)) - &
          4._wp*n*(5._wp - 3._wp*run))/(9._wp*xb)
        jll_nll(5,5,1) = (-2._wp*(21._wp - 4._wp*n*run))/(3._wp*xb)
        jll_nll(5,6,1) = -12._wp/xb
        jll_nll(5,4,2) = (2._wp*(9._wp - 2._wp*n*run))/(3._wp*xb)
        jll_nll(5,5,2) = 8._wp/xb
      end if
      if (order(6)) then
        jll_nll(6,4,1)=(2._wp*(-((243._wp - 8._wp*n**2)*run) + &
          3._wp*(n*(-4._wp*pi2 - 3._wp*(11._wp - run)) + &
          pi2*(27._wp - 8._wp*n*run))) + 27._wp*(9._wp - 40._wp*zeta3))/(27._wp*xb)
        jll_nll(6, 5, 1) = -1._wp/9._wp*(3._wp*(51._wp - 40._wp*pi2) + &
          8._wp*(n*(20._wp - 21._wp*run) + (27._wp + 2._wp*n**2)*run))/xb
        jll_nll(6,6,1) = (-4._wp*(15._wp - 4._wp*n*run))/xb
        jll_nll(6,7,1) = -32._wp/xb
        jll_nll(6,4,2) = (9._wp*(27._wp - 8._wp*pi2) - &
          16._wp*(9._wp - n)*n*run)/(18._wp*xb)
        jll_nll(6,5,2) = (4._wp*(27._wp - 8._wp*n*run))/(3._wp*xb)
        jll_nll(6,6,2) = 24._wp/xb
      end if
    else
      if (order(1)) then
        jll_nll(1,4,1)=2._wp*(-1._wp + xb**(-1))
      end if
      if (order(2)) then
        jll_nll(2,4,1)=2._wp*(-1._wp + 3._wp/xb)
        jll_nll(2,5,1)=8._wp*(-1._wp + xb**(-1))
      end if
      if (order(3)) then
        jll_nll(3,4,1)=(9._wp + 8._wp*pi2 + (27._wp - 8._wp*pi2)/xb)/2._wp
        jll_nll(3,5,1)=12._wp*(-1._wp + 3._wp/xb)
        jll_nll(3,6,1)=24._wp*(-1._wp + xb**(-1))
      end if
      if (order(4)) then
        jll_nll(4,4,1)=2._wp*(1._wp - xb**(-1))
        jll_nll(4,5,1)=4._wp*(1._wp - xb**(-1))
        jll_nll(4,4,2)=2._wp*(-1._wp + xb**(-1))
      end if
      if (order(5)) then
        jll_nll(5,4,1)=(-3._wp*(2._wp*(3._wp + 2._wp*pi2) - 9._wp*run) + 4._wp*n*(8._wp - &
          3._wp*run))/9._wp + (3._wp*(4._wp*pi2 + 3._wp*(1._wp - 3._wp*run)) - &
          4._wp*n*(5._wp - 3._wp*run))/(9._wp*xb)
        jll_nll(5,5,1) = (2._wp*(15._wp - 4._wp*n*run))/3._wp - (2._wp*(21._wp - 4._wp*n*run))/(3._wp*xb)
        jll_nll(5,6,1) = 12._wp - 12._wp/xb
        jll_nll(5,4,2) = (-2._wp*(3._wp - 2._wp*n*run))/3._wp + (2._wp*(9._wp - 2._wp*n*run))/(3._wp*xb)
        jll_nll(5,5,2) = -8._wp + 8._wp/xb
      end if
      if (order(6)) then
        jll_nll(6,4,1)=(2._wp*n*(-20._wp*n*run + 3._wp*(11._wp + 4._wp*(3._wp*run + &
          pi2*(1._wp + 2._wp*run)))))/27._wp + (135._wp + 2._wp*(run*(-2._wp*(81._wp - &
          4._wp*n**2) - 81._wp*x) + 3._wp*(n*(-4._wp*pi2 - 3._wp*(11._wp - run)) + &
        pi2*(4._wp*(3._wp - 2._wp*n*run) + 15._wp*x) + 18._wp*x*(1._wp - &
          10._wp*zeta3))))/(27._wp*xb)
        jll_nll(6,5,1)=(-3._wp*(21._wp + 8._wp*(5._wp*pi2 - 9._wp*run)) + &
          16._wp*n*(13._wp - (6._wp - n)*run))/9._wp - (3._wp*(51._wp - 40._wp*pi2) + &
          8._wp*(n*(20._wp - 21._wp*run) + (27._wp + 2._wp*n**2)*run))/(9._wp*xb)
        jll_nll(6,6,1) = 4._wp*(9._wp - 4._wp*n*run) - (4._wp*(15._wp - 4._wp*n*run))/xb
        jll_nll(6,7,1) = 32._wp - 32._wp/xb
        jll_nll(6,4,2)=(9._wp*(9._wp + 8._wp*pi2) + 16._wp*(3._wp - n)*n*run)/18._wp + &
          (9._wp*(27._wp - 8._wp*pi2) - 16._wp*(9._wp - n)*n*run)/(18._wp*xb)
        jll_nll(6,5,2) = (-4._wp*(9._wp - 8._wp*n*run))/3._wp + (4._wp*(27._wp - 8._wp*n*run))/(3._wp*xb)
        jll_nll(6,6,2) = -24._wp + 24._wp/xb
      end if
    end if
  end subroutine recbar_ele

  module subroutine recbar_pos (jll_nll, n, run, order)
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    jll_nll = 0._wp
  end subroutine recbar_pos

  module subroutine rechat_singlet (x, xb, xp, pl, jll_nll, n, run, order)
    real(wp), intent(in) :: x, xb, xp
    real(wp), dimension(12), intent(in) :: pl
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    real(wp), dimension(7,7,2) :: jll_nll_pos, jll_nll_ele
    call rechat_ele (x, xb, xp, pl, jll_nll_ele, n, run, order)
    call rechat_pos (x, xb, xp, pl, jll_nll_pos, n, run, order)
    jll_nll = add_logvec(jll_nll_ele,jll_nll_pos)
  end subroutine rechat_singlet

  module subroutine rechat_nonsinglet (x, xb, xp, pl, jll_nll, n, run, order)
    real(wp), intent(in) :: x, xb, xp
    real(wp), dimension(12), intent(in) :: pl
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    real(wp), dimension(7,7,2) :: jll_nll_pos, jll_nll_ele
    call rechat_ele (x, xb, xp, pl, jll_nll_ele, n, run, order)
    call rechat_pos (x, xb, xp, pl, jll_nll_pos, n, run, order)
    jll_nll = sub_logvec(jll_nll_ele,jll_nll_pos)
  end subroutine rechat_nonsinglet

  module subroutine rechat_photon (x, xb, xp, pl, jll_nll, n, run, order)
    real(wp), intent(in) :: x, xb, xp
    real(wp), dimension(12), intent(in) :: pl
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    jll_nll = 0._wp
    if (x .le. 1.0e-7_wp) then
      if (order(1)) then
        jll_nll(1,4,1) = 2._wp/x
      end if
      if (order(2)) then
        jll_nll(2,4,1) = (-4._wp*n)/(3._wp*x)
      end if
      if (order(3)) then
        jll_nll(3,3,1) = (-16._wp*n)/(3._wp*x)
        jll_nll(3,4,1) = (-4._wp*(31._wp - 2._wp*n)*n)/(9._wp*x)
      end if
      if (order(4)) then
        jll_nll(4,3,1) = -4._wp/x
        jll_nll(4,4,1) = -2._wp/x
        jll_nll(4,4,2) = 2._wp/x
      end if
      if (order(5)) then
        jll_nll(5,3,1) = (8._wp*n*(1._wp + run))/(3._wp*x)
        jll_nll(5,4,1) = (-4._wp*n*(7._wp - 3._wp*run) - 27._wp*run)/(9._wp*x)
        jll_nll(5,4,2) = (-4._wp*n*(1._wp + run))/(3._wp*x)
      end if
      if (order(6)) then
        jll_nll(6,2,1) = (16._wp*n)/(3._wp*x)
        jll_nll(6,3,1) = (8._wp*n*(15._wp - 2._wp*n*(1._wp + 3._wp*run)))/(9._wp*x)
        jll_nll(6,4,1)=(2._wp*n*(81._wp*run - 4._wp*(34._wp - n*(7._wp + run))) + &
          27._wp*(13._wp + 2._wp*(-pi2 + 8._wp*zeta3)))/(27._wp*x)
        jll_nll(6,3,2) = (-16._wp*n)/(3._wp*x)
        jll_nll(6,4,2) = (-4._wp*n*(31._wp - 2._wp*n*(1._wp + 3._wp*run)))/(9._wp*x)
      end if
    elseif (xb .le. 1.0e-7_wp) then
      if (order(1)) then
        jll_nll(1,4,1) = xb
      end if
      if (order(2)) then
        jll_nll(2,4,1) = -1._wp/6._wp*((3._wp + 4._wp*n)*xb)
        jll_nll(2,5,1) = 2._wp*xb
      end if
      if (order(3)) then
        jll_nll(3,4,1) = ((4._wp*n*(3._wp + 4._wp*n) + 3._wp*(99._wp - 8._wp*pi2))*xb)/36._wp
        jll_nll(3,5,1) = (-2._wp*(3._wp + 2._wp*n)*xb)/3._wp
        jll_nll(3,6,1) = 4._wp*xb
      end if
      if (order(4)) then
        jll_nll(4,4,1) = xb
        jll_nll(4,4,2) = xb
      end if
      if (order(5)) then
        jll_nll(5,4,1) = ((-9._wp*(4._wp + 3._wp*run) - 4._wp*n*(7._wp + 3._wp*run))*xb)/18._wp
        jll_nll(5,5,1) = ((3._wp - 4._wp*n)*xb)/3._wp
        jll_nll(5,6,1) = -3._wp*xb
        jll_nll(5,4,2) = -1._wp/6._wp*((3._wp + 4._wp*n*(1._wp + run))*xb)
        jll_nll(5,5,2) = 2._wp*xb
      end if
      if (order(6)) then
        jll_nll(6,4,1)=(xb*(4._wp*(n*(8._wp*n*(7._wp + 13._wp*run) - &
          3._wp*(21._wp - 2._wp*(2._wp*pi2 + 39._wp*run))) + &
          36._wp*ln2*(-4._wp*pi2 - (9._wp - 10._wp*ln2)*ln2)) + &
          27._wp*(35._wp - 4._wp*(pi2 - 3._wp*run - 32._wp*zeta3))))/216._wp
        jll_nll(6,5,1) = ((-3._wp*(99._wp - 4._wp*(pi2 - 9._wp*run)) + &
          4._wp*n*(4._wp*n*(1._wp + run) - 9._wp*(2._wp + run)))*xb)/18._wp
        jll_nll(6,6,1) = ((21._wp - 4._wp*n*(1._wp - 5._wp*run))*xb)/6._wp
        jll_nll(6,7,1) = -2._wp*xb
        jll_nll(6,4,2) = ((3._wp*(99._wp - 8._wp*pi2) + &
          4._wp*n*(3._wp*(1._wp + 2._wp*run) + 4._wp*n*(1._wp + 3._wp*run)))*xb)/36._wp
        jll_nll(6,5,2) = (-2._wp*(3._wp + 2._wp*n*(1._wp + 2._wp*run))*xb)/3._wp
        jll_nll(6,6,2) = 4._wp*xb
      end if
    else
      if (xb .le. 0.0_wp) then
        jll_nll = 0._wp
      elseif (x .ge. 1.0_wp) then
        jll_nll = 0._wp
      else
        if (order(1)) then
          jll_nll(1,4,1)=-3._wp + 2._wp/x + x
        end if
        if (order(2)) then
          jll_nll(2,3,1)=2._wp - x
          jll_nll(2,4,1)=(3._wp + 4._wp*n*(3._wp - 2._wp/x) + &
             (-3._wp - 4._wp*n)*x)/6._wp
          jll_nll(2,5,1)=2._wp*(-3._wp + 2._wp/x + x)
        end if
        if (order(3)) then
          jll_nll(3,2,1) = -1._wp/2._wp*((1._wp - 4._wp*n)*(2._wp - x))
          jll_nll(3,3,1) = (-4._wp*n*(1._wp - 5._wp*x) - 3._wp*(5._wp - 2._wp*x))/3._wp - &
            (16._wp*n)/(3._wp*x)
          jll_nll(3,4,1)=(4._wp*(-((31._wp - 2._wp*n)*n) + 3._wp*(-pi2 + &
            6._wp*pl(4))))/(9._wp*x) + (4._wp*n*(283._wp - 159._wp*x + 4._wp*(-(n*(3._wp - &
            x)) - 4._wp*xb*xp)) + 3._wp*(8._wp*pi2*(5._wp - 2._wp*x) + 3._wp*(-11._wp*xb - &
            16._wp*(2._wp - x)*(pl(2) + pl(4)))))/36._wp
          jll_nll(3,5,1) = (-8._wp*n)/(3._wp*x) + (2._wp*(2._wp*n*(3._wp - x) + 3._wp*xb))/3._wp
          jll_nll(3,6,1) = -4._wp*(3._wp - x) + 8._wp/x
        end if
        if (order(4)) then
          jll_nll(4,3,1)=2._wp*(2._wp*(1._wp - x**(-1)) - x)
          jll_nll(4,4,1)=3._wp - 2._wp/x - x
          jll_nll(4,4,2)=-3._wp + 2._wp/x + x
        end if
        if (order(5)) then
          jll_nll(5,2,1) = (-2._wp + x)/2._wp
          jll_nll(5,3,1) = (8._wp*n*(1._wp + run))/(3._wp*x) + (-8._wp*n*(1._wp + &
            run)*(2._wp - x) + 15._wp*x)/6._wp
          jll_nll(5,4,1)=(-4._wp*n*(7._wp - 3._wp*run) - &
            27._wp*run)/(9._wp*x) + (3._wp*(-2._wp*pi2*(2._wp - x) + &
            3._wp*(3._wp*run*(3._wp - x) + 5._wp*xb)) + 4._wp*(n*(3._wp*(9._wp - run*(3._wp - &
            x)) - 13._wp*x) + 9._wp*(2._wp - x)*pl(2)))/18._wp
          jll_nll(5,5,1) = (3._wp*(7._wp - 3._wp*x) + 4._wp*n*(3._wp - x))/3._wp - &
            (4._wp*(3._wp + 2._wp*n))/(3._wp*x)
          jll_nll(5,6,1) = 3._wp*(3._wp - x) - 6._wp/x
          jll_nll(5,3,2) = 2._wp - x
          jll_nll(5,4,2) = (-4._wp*n*(1._wp + run))/(3._wp*x) + (4._wp*n*(1._wp + &
            run)*(3._wp - x) + 3._wp*xb)/6._wp
          jll_nll(5,5,2) = -2._wp*(3._wp - x) + 4._wp/x
        end if
        if (order(6)) then
          jll_nll(6,1,1) = (2._wp*(1._wp - 4._wp*n*(2._wp - x)) + x)/3._wp
          jll_nll(6,2,1) = (16._wp*n)/(3._wp*x) + (-(n*(-(run*(2._wp - x)) + &
            12._wp*(2._wp + x))) + 6._wp*xp)/3._wp
          jll_nll(6,3,1)=(-2._wp*(-4._wp*n*(15._wp - 2._wp*n*(1._wp + 3._wp*run)) + &
            3._wp*(-pi2 + 6._wp*(pl(6) + pl(8)))))/(9._wp*x) + (3._wp*(2._wp*(81._wp + &
            2._wp*(4._wp*pi2 - 9._wp*run*(2._wp - x))) + 39._wp*x) + &
            4._wp*(n*(3._wp*(run*(4._wp - 3._wp*x) - 15._wp*x) - 2._wp*(397._wp - &
            4._wp*(n*(1._wp + 3._wp*run)*(2._wp - x) + 4._wp*xb*xp))) + 18._wp*(-((6._wp - &
            x)*(pl(6) + pl(2))) - (2._wp + x)*pl(8))))/36._wp
          jll_nll(6,4,1)=(4._wp*(-(n*(2993._wp - 3._wp*(1179._wp - 8._wp*pi2*run)*x - &
            4._wp*(3._wp*(-9._wp*run*(5._wp - 2._wp*x) - pi2*(3._wp - 4._wp*run - x)) + &
            2._wp*(-(n*(27._wp + run*(9._wp - 7._wp*x) - 13._wp*x)) - 14._wp*xb*xp)))) + &
            9._wp*(pi2*(24._wp*ln2 + (2._wp - 3._wp*x)*pl(1)) + pl(1)*((18._wp + &
            5._wp*x)*pl(1)**2 + 6._wp*(10._wp + x)*pl(6)) + 2._wp*((-8._wp*n*run*(2._wp - x) - &
            9._wp*(2._wp + x))*pl(2) + (3._wp*(4._wp - 7._wp*x) + 4._wp*n*(2._wp - &
            x))*pl(4) + 3._wp*(-4._wp*(5._wp + 2._wp*x) + (6._wp - x)*pl(1))*pl(8) + &
            2._wp*(-3._wp*((5._wp + 2._wp*x)*pl(1)**2 - (14._wp - 3._wp*x)*pl(3) - (10._wp - &
            3._wp*x)*pl(5)) + 2._wp*(-7._wp*ln2**3 + 3._wp*((6._wp - x)*pl(7) + 2._wp*(2._wp &
            - x)*pl(9))))))) - 27._wp*(135._wp - 4._wp*(3._wp*(pi2 - run) - &
            56._wp*zeta3)))/216._wp + (27._wp*(135._wp - xb*xp*(31._wp + 32._wp*zeta3)) + &
            4._wp*(81._wp*run*(1._wp - xb*xp) + 2._wp*(9._wp*(pi2*(2._wp*(5._wp - &
            3._wp*xb*xp) - 7._wp*pl(1)) + pl(1)*(-((18._wp - 17._wp*pl(1))*pl(1)) + &
            6._wp*(pl(6) - pl(8)))) + 2._wp*(n*(81._wp*run + 2._wp*(3._wp*pi2 - &
            2._wp*(34._wp - n*(7._wp + run) + 9._wp*pl(4)))) + 54._wp*(-3._wp*pl(8) + &
            2._wp*(-pl(4) + pl(7)) + 5._wp*zeta3)))))/(216._wp*x)
          jll_nll(6,5,1)=-1._wp/9._wp*(45._wp - 4._wp*(3._wp*(-(n*(8._wp - run)) - &
            9._wp*run) + 2._wp*(2._wp*n**2*(1._wp + run) + 3._wp*(pi2 - 3._wp*pl(4)))))/x + &
            (3._wp*(3._wp*(3._wp*(11._wp + 4._wp*run*(3._wp - x)) - 23._wp*x) - pi2*(42._wp &
            - 19._wp*x)) + 2._wp*(2._wp*n*(3._wp*(2._wp*(13._wp - 5._wp*x) - run*(3._wp - &
            x)) - 4._wp*n*(1._wp + run)*(3._wp - x)) - 9._wp*(2._wp - x)*(-7._wp*pl(2) - &
            pl(4))))/18._wp
          jll_nll(6,6,1)=(3._wp*(25._wp - 9._wp*x) + 4._wp*n*(1._wp - 5._wp*run)*(3._wp - &
            x))/6._wp - (4._wp*(6._wp + n*(1._wp - 5._wp*run)))/(3._wp*x)
          jll_nll(6,7,1) = 2._wp*(7._wp - 3._wp*x) - 8._wp/x
          jll_nll(6,2,2) = -1._wp/2._wp*((1._wp - 4._wp*n)*(2._wp - x))
          jll_nll(6,3,2) = (-4._wp*n*(1._wp + run*(2._wp - x) - 5._wp*x) - 3._wp*(5._wp &
            - 2._wp*x))/3._wp - (16._wp*n)/(3._wp*x)
          jll_nll(6,4,2)=(4._wp*(-(n*(31._wp - 2._wp*n*(1._wp + 3._wp*run))) + 3._wp*(-pi2 &
            + 6._wp*pl(4))))/(9._wp*x) + (4._wp*n*(283._wp - 159._wp*x + &
            2._wp*(-3._wp*run*xb + 2._wp*(-(n*(1._wp + 3._wp*run)*(3._wp - x)) - &
            4._wp*xb*xp))) + 3._wp*(8._wp*pi2*(5._wp - 2._wp*x) + 3._wp*(-11._wp*xb - &
            16._wp*(2._wp - x)*(pl(2) + pl(4)))))/36._wp
          jll_nll(6,5,2) = (-8._wp*n*(1._wp + 2._wp*run))/(3._wp*x) + &
            (2._wp*(2._wp*n*(1._wp + 2._wp*run)*(3._wp - x) + 3._wp*xb))/3._wp
          jll_nll(6,6,2) = -4._wp*(3._wp - x) + 8._wp/x
        end if
      end if
    end if
  end subroutine rechat_photon

  module subroutine rechat_ele (x, xb, xp, pl, jll_nll, n, run, order)
    real(wp), intent(in) :: x, xb, xp
    real(wp), dimension(12), intent(in) :: pl
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    jll_nll=0._wp
    if (x .le. 1.0e-7_wp) then
      if (order(1)) then
        jll_nll(1,4,1) = 1._wp
      end if
      if (order(2)) then
        jll_nll(2,4,1) = (4._wp*n)/(3._wp*x)
      end if
      if (order(3)) then
        jll_nll(3,4,1) = (-8._wp*n**2)/(9._wp*x)
      end if
      if (order(4)) then
        jll_nll(4,4,1) = -1._wp
        jll_nll(4,4,2) = 1._wp
      end if
      if (order(5)) then
        jll_nll(5,3,1) = (-8._wp*n)/(3._wp*x)
        jll_nll(5,4,1) = (-2._wp*n)/x
        jll_nll(5,4,2) = (4._wp*n)/(3._wp*x)
      end if
      if (order(6)) then
        jll_nll(6,3,1) = (16._wp*n**2*(1._wp + 2._wp*run))/(9._wp*x)
        jll_nll(6,4,1) = (4._wp*n*(-(n*(1._wp - 28._wp*run)) + &
          3._wp*(2._wp*(6._wp - pi2) - 9._wp*run)))/(27._wp*x)
        jll_nll(6,4,2) = (-8._wp*n**2*(1._wp + 2._wp*run))/(9._wp*x)
      end if
    elseif (xb .le. 1.0e-7_wp) then
      if (order(1)) then
        jll_nll(1,4,1) = xb
      end if
      if (order(2)) then
        jll_nll(2,4,1) = -((3._wp - n)*xb)
        jll_nll(2,5,1) = 4._wp*xb
      end if
      if (order(3)) then
        jll_nll(3,4,1) = ((-4._wp*n*(3._wp + 2._wp*n) + &
          3._wp*(31._wp - 8._wp*pi2))*xb)/12._wp
        jll_nll(3,5,1) = -2._wp*(9._wp - 2._wp*n)*xb
        jll_nll(3,6,1) = 12._wp*xb
      end if
      if (order(4)) then
        jll_nll(4,4,1) = -xb
        jll_nll(4,5,1) = -2._wp*xb
        jll_nll(4,4,2) = xb
      end if
      if (order(5)) then
        jll_nll(5,4,1) = ((3._wp*(-2._wp*(9._wp - 2._wp*pi2) - 9._wp*run) - &
          2._wp*n*(37._wp - 6._wp*run))*xb)/18._wp
        jll_nll(5,5,1) = ((9._wp + 4._wp*n*run)*xb)/3._wp
        jll_nll(5,6,1) = -6._wp*xb
        jll_nll(5,4,2) = -1._wp/3._wp*((9._wp - n*(3._wp - 2._wp*run))*xb)
        jll_nll(5,5,2) = 4._wp*xb
      end if
      if (order(6)) then
        jll_nll(6,4,1)=(xb*(n*(-2._wp*n*(21._wp - 40._wp*run) - &
          3._wp*(-3._wp*(77._wp + 3._wp*run) + pi2*(7._wp + 8._wp*run))) - &
          9._wp*(-27._wp*run + 10._wp*(pi2 - 3._wp*(5._wp - 2._wp*zeta3)))))/27._wp
        jll_nll(6,5,1)=((-3._wp*(195._wp - 8._wp*(5._wp*pi2 - 9._wp*run)) - &
          2._wp*n*(197._wp - 4._wp*(-12._wp*run - n*(3._wp + 2._wp*run))))*xb)/18._wp
        jll_nll(6,6,1) = 2._wp*(11._wp - n*(1._wp - 4._wp*run))*xb
        jll_nll(6,7,1) = -16._wp*xb
        jll_nll(6,4,2) = ((9._wp*(31._wp - 8._wp*pi2) + 4._wp*n*(-9._wp*(1._wp - 4._wp*run) - &
          2._wp*n*(3._wp + 4._wp*run)))*xb)/36._wp
        jll_nll(6,5,2) = (-2._wp*(27._wp - 2._wp*n*(3._wp - 4._wp*run))*xb)/3._wp
        jll_nll(6,6,2) = 12._wp*xb
      end if
    else
      if (order(1)) then
        jll_nll(1,4,1)=xb
      end if
      if (order(2)) then
        jll_nll(2,3,1) = -4._wp/xb + (3._wp + 2._wp*n)*xp
        jll_nll(2,4,1) = (4._wp*n)/(3._wp*x) + &
          (-3._wp*(3._wp + x) - n*(1._wp + 3._wp*x - 4._wp*xb*xp))/3._wp
        jll_nll(2,5,1) = 4._wp*xb
      end if
      if (order(3)) then
        jll_nll(3,2,1) = 4._wp/xb - ((7._wp + 4._wp*n)*xp)/2._wp
        jll_nll(3,3,1) = -18._wp/xb + (9._wp*(11._wp + 3._wp*x) + 8._wp*n*(4._wp + 3._wp*x + (-n - &
          4._wp*xb)*xp))/6._wp
        jll_nll(3,4,1)=(-8._wp*n**2)/(9._wp*x) - (4._wp*(pi2 - 6._wp*(pl(2) + pl(4))))/xb + &
          (4._wp*(n*(-3._wp*(23._wp - 23._wp*x - 4._wp*pi2*xp) + 2._wp*n*(1._wp + 3._wp*x - &
          4._wp*xb*xp)) - 18._wp*(9._wp + 4._wp*n)*xp*pl(2)) + 9._wp*(4._wp*pi2*(1._wp + 5._wp*x) - &
          3._wp*(19._wp + 5._wp*x + 16._wp*xp*pl(4))))/36._wp
        jll_nll(3,5,1) = (16._wp*n)/(3._wp*x) + (2._wp*(-9._wp*(3._wp + x) - 2._wp*n*(1._wp + &
          3._wp*x - 4._wp*xb*xp)))/3._wp
        jll_nll(3,6,1) = 12._wp*xb
      end if
      if (order(4)) then
        jll_nll(4,4,1)=-xb
        jll_nll(4,4,2)=xb
        jll_nll(4,5,1)=-2._wp*xb
      end if
      if (order(5)) then
        jll_nll(5,2,1) = -1._wp/2._wp*((1._wp + 6._wp*n)*xp)
        jll_nll(5,3,1) = (-8._wp*n)/(3._wp*x) + (3._wp - 4._wp*n)/(3._wp*xb) + &
          (-(n*(13._wp + 8._wp*x)*xb) - 9._wp*xp)/3._wp
        jll_nll(5,4,1)=(-2._wp*n)/x + (2._wp*(pi2 - 6._wp*(pl(2) + pl(4))))/(3._wp*xb) + &
          (3._wp*((27._wp - 8._wp*pi2)*x - 3._wp*(7._wp + 3._wp*run*xb)) + &
          2._wp*(-(n*(79._wp - 85._wp*x - 6._wp*xb*(run + 3._wp*xp))) + &
          18._wp*xp*(2._wp*pl(2) + pl(4))))/18._wp
        jll_nll(5,5,1) = (3._wp*(1._wp + 3._wp*x) + 4._wp*n*run*xb)/3._wp
        jll_nll(5,6,1) = -6._wp*xb
        jll_nll(5,3,2) = -4._wp/xb + (3._wp + 2._wp*n)*xp
        jll_nll(5,4,2) = (4._wp*n)/(3._wp*x) + (-3._wp*(3._wp + x) - n*(1._wp + 3._wp*x - &
          2._wp*xb*(-run + 2._wp*xp)))/3._wp
        jll_nll(5,5,2) = 4._wp*xb
      end if
      if (order(6)) then
        jll_nll(6,1,1) = ((1._wp + 6._wp*n)*xp)/3._wp
        jll_nll(6,2,1)=(2._wp*(3._wp + 4._wp*n))/(3._wp*xb) + (9._wp*x + &
          n*(-5._wp*(4._wp + 3._wp*x) + (run + 2._wp*(n*(2._wp + 5._wp*run) + &
          10._wp*xb))*xp))/3._wp
        jll_nll(6,3,1)=(16._wp*n**2*(1._wp + 2._wp*run))/(9._wp*x) + &
          (2._wp*(3._wp*(3._wp - 2._wp*pi2) + n*(22._wp - 15._wp*run) + 2._wp*(27._wp + &
          2._wp*n**2)*run))/(9._wp*xb) + (n*(277._wp + 2._wp*(3._wp*(-(pi2*(7._wp + &
          4._wp*x)) - run*(3._wp + 5._wp*x)) + 8._wp*(-9._wp*x - 4._wp*xb*xp) + &
          n*(run*(23._wp + 7._wp*x + 8._wp*xb*xp) - 2._wp*(11._wp - 4._wp*(-4._wp*x + &
          xb*xp))))) + 3._wp*(3._wp*(7._wp - 9._wp*(x + run*xp)) + &
          2._wp*(3._wp*(-2._wp*n*(2._wp - x) - 3._wp*xp)*pl(2) + xp*(2._wp*pi2 - &
          3._wp*pl(4)))))/9._wp
        jll_nll(6,4,1)=(-2._wp*n*(9._wp*(3._wp*(1._wp - run*(1._wp - 3._wp/xb)) - &
          (11._wp - 2._wp*pi2)/xb) + 2._wp*(n*(1._wp - 28._wp*run) + 18._wp*(-2._wp*pl(2) &
          - pl(4)))))/(27._wp*x) - (3._wp*(-(pi2*(5._wp - 4._wp*(n*run - x))) + &
          3._wp*(-41._wp*xb + 6._wp*(3._wp*pl(2) + 4._wp*pl(4)))) + &
          2._wp*(9._wp*run*(-3._wp*xb - 4._wp*n*(pl(2) + pl(4))) + n*(3._wp*(11._wp - &
          9._wp*run) + 2._wp*(-pi2 - 12._wp*(pl(2) + pl(4))))))/(9._wp*xb) + &
          (4._wp*(9._wp*((6._wp*(7._wp + 3._wp*x) + n*(10._wp*(1._wp - run - run*x) - &
          (1._wp - 4._wp*(n + 2._wp*xb))*xp))*pl(2) + (6._wp*(5._wp + x) - n*(5._wp + &
          7._wp*x + 2._wp*(3._wp*run - 2._wp*xb)*xp))*pl(4) + 3._wp*((4._wp*n*(4._wp + x) &
          + 7._wp*xp)*pl(3) + 2._wp*(n*(5._wp + 2._wp*x) + 6._wp*xp)*pl(5))) + &
          n*(-(n*(7._wp - 9._wp*x + 2._wp*((19._wp + 3._wp*pi2 - 19._wp*x)*xp + &
          2._wp*run*(-11._wp*(1._wp - 2._wp*x) - 5._wp*xb*xp)))) + 3._wp*(83._wp - &
          85._wp*x + pi2*(run + 2._wp*(2._wp*x - 3._wp*(1._wp + xb*xp))) + &
          3._wp*(run*(-((1._wp - 3._wp*pi2)*x) + 2._wp*(1._wp - 3._wp*xb*xp)) + &
          3._wp*(xp*(xb - 4._wp*zeta3) - 12._wp*zeta3))))) - 9._wp*((99._wp + &
          2._wp*(10._wp*pi2 - 9._wp*run))*xp + 12._wp*(17._wp - 3._wp*x)*zeta3))/54._wp
        jll_nll(6,5,1)=(-8._wp*n*(1._wp + 2._wp*n))/(9._wp*x) + (4._wp*(pi2 - &
          6._wp*(pl(2) + pl(4))))/xb + (-3._wp*(-3._wp*(5._wp + 31._wp*x) + &
          2._wp*(-(pi2*(9._wp - 31._wp*x)) + 36._wp*run*xb)) + 2._wp*(-(n*(333._wp - &
          293._wp*x - 2._wp*(-3._wp*pi2*xp + 2._wp*((-9._wp*run + 16._wp*xb)*xp + n*(1._wp &
          + 3._wp*x - 2._wp*xb*(run + 2._wp*xp)))))) + 18._wp*xp*((11._wp + 2._wp*n)*pl(2) &
          - (1._wp + 2._wp*n)*pl(4))))/18._wp
        jll_nll(6,6,1) = (-8._wp*n)/(3._wp*x) + (2._wp*(3._wp*(7._wp + 5._wp*x) + &
          n*(1._wp + 3._wp*x - 4._wp*xb*(-3._wp*run + xp))))/3._wp
        jll_nll(6,7,1) = -16._wp*xb
        jll_nll(6,2,2) = 4._wp/xb - ((7._wp + 4._wp*n)*xp)/2._wp
        jll_nll(6,3,2)=(-2._wp*(27._wp - 8._wp*n*run))/(3._wp*xb) + (9._wp*(11._wp + &
          3._wp*x) + 8._wp*n*(4._wp - (n*(1._wp + 2._wp*run) + 4._wp*xb)*xp - 3._wp*(-x + &
          run*xp)))/6._wp
        jll_nll(6,4,2)=(-8._wp*n**2*(1._wp + 2._wp*run))/(9._wp*x) - (4._wp*(pi2 - &
          6._wp*(pl(2) + pl(4))))/xb + (4._wp*(n*(2._wp*n*(1._wp + 3._wp*x + 4._wp*(run - &
          (1._wp + 2._wp*run)*xb)*xp) - 3._wp*(23._wp - 23._wp*x - 4._wp*(run*(3._wp + x) &
          + pi2*xp))) - 18._wp*(9._wp + 4._wp*n)*xp*pl(2)) + 9._wp*(4._wp*pi2*(1._wp + &
          5._wp*x) - 3._wp*(19._wp + 5._wp*x + 16._wp*xp*pl(4))))/36._wp
        jll_nll(6,5,2) = (16._wp*n)/(3._wp*x) + (2._wp*(-9._wp*(3._wp + x) - &
          2._wp*n*(1._wp + 3._wp*x - 4._wp*xb*(-run + xp))))/3._wp
        jll_nll(6,6,2) = 12._wp*xb
      end if
    end if
  end subroutine rechat_ele

  module subroutine rechat_pos (x, xb, xp, pl, jll_nll, n, run, order)
    real(wp), intent(in) :: x, xb, xp
    real(wp), dimension(12), intent(in) :: pl
    real(wp), intent(in) :: n, run
    logical, dimension(6), intent(in) :: order
    real(wp), dimension(7,7,2), intent(out) :: jll_nll
    jll_nll = 0._wp
    if (x .le. 1.0e-7_wp) then
      if (order(1)) then
        jll_nll(1,:,:)=0._wp
      end if
      if (order(2)) then
        jll_nll(2,4,1) = (4._wp*n)/(3._wp*x)
      end if
      if (order(3)) then
        jll_nll(3,4,1) = (-8._wp*n**2)/(9._wp*x)
      end if
      if (order(4)) then
        jll_nll(4,:,:)=0._wp
      end if
      if (order(5)) then
        jll_nll(5,3,1) = (-8._wp*n)/(3._wp*x)
        jll_nll(5,4,1) = (-2._wp*n)/x
        jll_nll(5,4,2) = (4._wp*n)/(3._wp*x)
      end if
      if (order(6)) then
        jll_nll(6,3,1) = (16._wp*n**2*(1._wp + 2._wp*run))/(9._wp*x)
        jll_nll(6,4,1) = (4._wp*n*(-(n*(1._wp - 28._wp*run)) + &
          3._wp*(2._wp*(6._wp - pi2) - 9._wp*run)))/(27._wp*x)
        jll_nll(6,4,2) = (-8._wp*n**2*(1._wp + 2._wp*run))/(9._wp*x)
      end if
    elseif (xb .le. 1.0e-7_wp) then
      if (order(1)) then
        jll_nll(1,:,:)=0._wp
      end if
      if (order(2)) then
        jll_nll(2,4,1) = n*xb
      end if
      if (order(3)) then
        jll_nll(3,4,1) = -1._wp/3._wp*(n*(3._wp + 2._wp*n)*xb)
        jll_nll(3,5,1) = 4._wp*n*xb
      end if
      if (order(4)) then
        jll_nll(4,:,:)=0._wp
      end if
      if (order(5)) then
        jll_nll(5,4,1) = -(n*xb)
        jll_nll(5,4,2) = n*xb
      end if
      if (order(6)) then
        jll_nll(6,4,1)=(xb*(2._wp*(6._wp*pi2 + n*(-3._wp*(pi2 - 3._wp*(14._wp - 3._wp*run)) - &
          2._wp*n*(7._wp - 6._wp*run))) - 45._wp*ln2*(6._wp + 5._wp*ln2)))/18._wp
        jll_nll(6,5,1) = ((-(n*(15._wp + 4._wp*n)) + pi2)*xb)/3._wp
        jll_nll(6,6,1) = -2._wp*n*xb
        jll_nll(6,4,2) = -1._wp/3._wp*(n*(3._wp + 2._wp*n*(1._wp + 2._wp*run))*xb)
        jll_nll(6,5,2) = 4._wp*n*xb
      end if
    else
      if (order(1)) then
        jll_nll(1,:,:)=0._wp
      end if
      if (order(2)) then
        jll_nll(2,3,1) = 2._wp*n*xp
        jll_nll(2,4,1) = (4._wp*n)/(3._wp*x) - (n*(1._wp + 3._wp*x - 4._wp*xb*xp))/3._wp
      end if
      if (order(3)) then
        jll_nll(3,2,1) = -2._wp*n*xp
        jll_nll(3,3,1) = (4._wp*n*(4._wp + 3._wp*x + (-n - 4._wp*xb)*xp))/3._wp
        jll_nll(3,4,1)=(-8._wp*n**2)/(9._wp*x) + (n*(-3._wp*(23._wp - 23._wp*x - &
          4._wp*pi2*xp) + 2._wp*(n*(1._wp + 3._wp*x - 4._wp*xb*xp) - &
          36._wp*xp*pl(2))))/9._wp
        jll_nll(3,5,1) = (16._wp*n)/(3._wp*x) - (4._wp*n*(1._wp + 3._wp*x - 4._wp*xb*xp))/3._wp
      end if
      if (order(4)) then
        jll_nll(4,:,:)=0._wp
      end if
      if (order(5)) then
        jll_nll(5,2,1) = -xb + 2._wp/xp - 3._wp*n*xp
        jll_nll(5,3,1) = (-8._wp*n)/(3._wp*x) + (6._wp*xp - n*(7._wp - 3._wp*x + &
          8._wp*xb*xp))/3._wp
        jll_nll(5,4,1)=(-2._wp*n)/x + (2._wp*(pi2 + 6._wp*(-pl(1)**2 - &
          2._wp*pl(8))))/(3._wp*xp) + (-3._wp*n*(5._wp - 7._wp*x - &
          2._wp*xb*xp) + xb*(-pi2 + 6._wp*(pl(1)**2 + 2._wp*(1._wp + pl(8)))))/3._wp
        jll_nll(5,3,2) = 2._wp*n*xp
        jll_nll(5,4,2) = (4._wp*n)/(3._wp*x) - (n*(1._wp + 3._wp*x - 4._wp*xb*xp))/3._wp
      end if
      if (order(6)) then
        jll_nll(6,1,1) = (2._wp*(xb + 3._wp*n*xp))/3._wp
        jll_nll(6,2,1)=(2._wp*(3._wp - 2._wp*run*n))/(3._wp*xp) + (-6._wp*(2._wp - x) + &
          n*(-9._wp*x - 2._wp*(7._wp - (2._wp + 5._wp*run)*n*xp - xb*(run + &
          10._wp*xp))))/3._wp
        jll_nll(6,3,1)=(16._wp*(1._wp + 2._wp*run)*n**2)/(9._wp*x) - (2._wp*(pi2 + &
          6._wp*(pl(2) + pl(4))))/(3._wp*xp) + (2._wp*(pi2 - 6._wp*(pl(6) + &
          pl(8))))/(3._wp*xb) + (n*(283._wp + 2._wp*(-33._wp*x - pi2*(21._wp + 12._wp*x) - &
          (32._wp + 32._wp*x)*xb + n*(-2._wp*(11._wp + 16._wp*x - 4._wp*xb*xp) + &
          run*(25._wp + 9._wp*x + 8._wp*xb*xp)) + 3._wp*(-11._wp*run*xp - &
          6._wp*(3._wp - xp)*pl(2)))) + 6._wp*((15._wp - pi2)*x - &
          3._wp*(3._wp - 2._wp*(pl(6) - (1._wp - 2._wp*x)*pl(8)))))/9._wp
        jll_nll(6,4,1)=(n*(-3._wp*(103._wp - 151._wp*xb*xp) + 4._wp*(-((1._wp - &
          28._wp*run)*n) + 9._wp*(-3._wp*run - pi2 + 2._wp*(2._wp*pl(2) + &
          pl(4))))))/(27._wp*x) - (pi2*(8._wp*ln2 - 5._wp*pl(1)) + &
          3._wp*pl(1)*(5._wp*pl(1)**2 + 2._wp*(-pl(6) + pl(8))) + 4._wp*(-5._wp*ln2**3 + &
          6._wp*(-pl(7) + 2._wp*pl(9))))/(3._wp*xb) - (pi2*(-15._wp*(5._wp + 3._wp*x) + &
          2._wp*(2._wp*(run*n + 3._wp*ln2) + 51._wp*pl(1))) + &
          6._wp*(pl(1)*(pl(1)*(4._wp*(9._wp - run*n) + 15._wp*pl(1)) + 6._wp*(5._wp*pl(6) - &
          4._wp*(pl(2) + pl(4)))) + 2._wp*(2._wp*(3._wp - 2._wp*run*n)*pl(8) + &
          15._wp*(-ln2**3 - pl(1)*pl(8) - 2._wp*(ln2*pl(10) + pl(11))))))/(9._wp*xp) - &
          (20._wp*(1._wp - 3._wp*x)*zeta3)/(xb*xp) + (-9._wp*(2._wp*(36._wp + &
          17._wp*pi2)*xb + 3._wp*(27._wp + 13._wp*x)*zeta3) + 2._wp*(9._wp*(pi2*(8._wp*ln2 &
          - (2._wp + 3._wp*x)*pl(1)) + pl(1)**2*(-4._wp*run*n*xb + 3._wp*(19._wp + 7._wp*x + &
          5._wp*(2._wp - x)*pl(1))) + 2._wp*(5._wp*(-3._wp*(1._wp + 2._wp*x) - &
          2._wp*ln2)*ln2**2 + (n*(13._wp + 3._wp*x + 8._wp*xb*xp) + 4._wp*(-((3._wp - &
          n**2)*xp) - 3._wp*xb*pl(1)))*pl(2) + (-(n*(1._wp + 3._wp*x - 4._wp*xb*xp)) - &
          12._wp*xb*pl(1))*pl(4) + (4._wp*(3._wp*(1._wp - 2._wp*x) - run*n*xb) + &
          3._wp*(4._wp - 3._wp*x)*pl(1))*pl(8) + 3._wp*((10._wp*x + (6._wp - &
          7._wp*x)*pl(1))*pl(6) + 2._wp*(-2._wp*pl(7) + (2._wp*n*(4._wp + x) + xb)*pl(3) + &
          (n*(5._wp + 2._wp*x) - 2._wp*xb)*pl(5) + (1._wp + 3._wp*x)*pl(9) + xb*pl(12))))) &
          + n*(-2._wp*n*(7._wp - 9._wp*x - 2._wp*(-((19._wp + 3._wp*pi2 - 19._wp*x)*xp) - &
          2._wp*run*(-2._wp*(2._wp - 9._wp*x) - 5._wp*xb*xp))) + 3._wp*(103._wp - &
          run*(-2._wp*pi2*xb + 3._wp*(5._wp - 17._wp*x + 12._wp*xb*xp)) - &
          6._wp*(-3._wp*xp*(xb - 4._wp*zeta3) + 2._wp*(pi2*(1._wp + xb*xp) + &
          18._wp*zeta3))))))/54._wp
        jll_nll(6,5,1)=(-8._wp*n*(1._wp + 2._wp*n))/(9._wp*x) - (4._wp*(15._wp*(ln2**2 - &
          pl(1)**2) + 2._wp*(pi2 + 6._wp*pl(6))))/(3._wp*xp) + (6._wp*(pi2*(3._wp - &
          2._wp*x) + 12._wp*xb*(2._wp + pl(6))) - n*(157._wp - 165._wp*x - &
          2._wp*(-3._wp*pi2*xp + 2._wp*(n*(1._wp + 3._wp*x - 4._wp*xb*xp) + xp*(16._wp*xb &
          + 9._wp*(pl(2) - pl(4)))))))/9._wp
        jll_nll(6,6,1) = (-8._wp*n)/(3._wp*x) + (2._wp*n*(1._wp + 3._wp*x - 4._wp*xb*xp))/3._wp
        jll_nll(6,2,2) = -2._wp*n*xp
        jll_nll(6,3,2) = (4._wp*n*(x*(3._wp + 4._wp*x) - (1._wp + 2._wp*run)*n*xp))/3._wp
        jll_nll(6,4,2)=(-8._wp*(1._wp + 2._wp*run)*n**2)/(9._wp*x) + (n*(2._wp*(1._wp + &
          2._wp*run)*n*(1._wp + 3._wp*x - 4._wp*xb*xp) - 3._wp*(23._wp - 23._wp*x - &
          4._wp*xp*(pi2 - 6._wp*pl(2)))))/9._wp
        jll_nll(6,5,2) = (16._wp*n)/(3._wp*x) - (4._wp*n*(1._wp + 3._wp*x - 4._wp*xb*xp))/3._wp
      end if
    end if
  end subroutine rechat_pos

  pure function sum_rm (log_xb, al0_2pi, ca, cb, d1, d2, cc, k0) result (s_rm)
    real(wp), intent(in) :: log_xb, al0_2pi, ca, cb, d1, d2, cc, k0
    real(wp),dimension(7,2) :: s_rm
    real(wp),dimension(5,7) :: mf, nf
    real(wp),dimension(5,2) :: rr, ss
    real(wp), parameter :: f0=1._wp-pi2/6._wp, f0ln=3._wp/4._wp
    real(wp) :: den1, den2, g1, g2, g3, g4, ta, tb, fac, logxb10
    integer :: i, j
    fac = exp(-(eulerc+d1-log_xb)*k0) / gamma(1._wp+k0)
    logxb10 = - log_xb / ln10
    nf = 0._wp
    nf(1,4) = 1._wp
    nf(2,4) = 1._wp
    nf(3,5) = -1._wp
    nf(4,4) = -pi2/6._wp
    nf(4,6) = 1._wp
    nf(5,4) = -2._wp*zeta3
    nf(5,5) = pi2/2._wp
    nf(5,7) = -1._wp
    mf = 0._wp
    mf(3,4) = (pi2/6._wp - zeta3*k0)*k0
    mf(4,4) = (2._wp*zeta3 - pi4/180._wp*k0)*k0
    mf(4,5) = -mf(3,4)*2._wp
    mf(5,4) = -(pi4/60._wp - 3._wp*k0*(pi2*zeta3/2._wp - 4._wp*zeta5))*k0
    mf(5,5) = -mf(4,4)*3._wp
    mf(5,6) = mf(3,4)*3._wp
    ta = ca + d1
    tb = f0 - ta*(1._wp + d1) - cb
    rr = 0._wp
    den1 = (d1 - log_xb)**(-1)
    den2 = (d2 - log_xb)**(-1)
    g1 = (fac*(1._wp - (mf(3,4) - (nf(4,4) + mf(4,4))*den1)*den1) - &
      1._wp - nf(4,4)*den1**2)*den1*al0_2pi*(cb + d1*ta)
    g2 = (fac*cc*(1._wp - (mf(3,4) - (nf(4,4) + mf(4,4))*den2)*den2) - &
      1._wp - nf(4,4)*den2**2)*den2
    rr(1,1) = (g1 - g2)*(0.5_wp + al0_2pi*(f0 + 0.25_wp)) - &
      al0_2pi*(g1*(0.5_wp + d1)**2 - g2*(0.5_wp + d2)**2)
    rr(1,2) = al0_2pi*((g1 - g2)*f0ln - (g1*d1 - g2*d2))
    rr(2,1) = -ta/2._wp - cc*(1._wp + d2) - al0_2pi*(ca*f0 + d1*tb - cb)
    rr(2,2) = -cc - al0_2pi*(cb + ta*f0ln + d1*(d1 - ca))
    rr(3,1) = 0.5_wp + cc + al0_2pi*tb
    rr(3,2) = -al0_2pi*(ta + 2._wp*f0ln)
    rr(4,1) = 1._wp + ta
    rr(4,2) = 1._wp
    rr(5,1) = - rr(4,2)
    ss = 0._wp
    ss(2,1) = (cc - 1._wp)*(1._wp + d2)
    ss(2,2) = -1._wp + cc
    ss(3,1) = -ss(2,2)
    s_rm = 0._wp
    do i = 1, size(s_rm,1)
      do j = 1, size(s_rm,2)
        s_rm(i,j) = rr(1,j)*nf(1,i) + &
          al0_2pi*( &
            ((fac - 1._wp)*rr(2,j) - ss(2,j))*nf(2,i) + &
            ((fac - 1._wp)*nf(3,i) + fac*mf(3,i))*rr(3,j) - ss(3,j)*nf(3,i) &
          )
        s_rm(i,j) = s_rm(i,j) + al0_2pi**2*( &
            (fac - 1._wp)*(nf(4,i)*rr(4,j) + nf(5,i)*rr(5,j)) + &
            fac*(mf(4,i)*rr(4,j) + mf(5,i)*rr(5,j)) &
          )
      end do
    end do
  end function sum_rm

  module function t_alpha (epdf, scale) result (t)
    real(wp) :: t
    type(qed_pdf_t), intent(in) :: epdf
    real(wp), intent(in) :: scale
    real(wp) :: alphamu, alpharef
    select type (alpha => epdf%aqed)
    type is (alpha_qed_from_scale_t)
       alpharef = real(alpha%ref,kind=wp)
       alphamu =  alpha%get(scale)
    type is (alpha_qed_fixed_t)
       call msg_fatal &
            ("t integrator: has to be called with running alpha.")
    end select
    t = log(alphamu/alpharef)/(2._wp*coeffqed_b0(0, epdf%nlep))
  end function t_alpha

  module pure function full_series ( &
      log_x, log_xb, ln0, p, al_2pi, a01, a02, a03, a04, a05) result (res)
    real(wp) :: res
    real(wp), intent(in) :: log_x, log_xb, p, al_2pi, ln0
    real(wp), intent(in), dimension(7,7,2) :: a01
    real(wp), intent(in), dimension(7,7,2), optional :: a02, a03, a04, a05
    real(wp), dimension(5,7,7,2) :: aa
    real(wp), dimension(7) :: fac
    real(wp), dimension(7) :: lv
    real(wp), dimension(2) :: l0v
    real(wp) :: temp, s, t, c , cc, cs, ccs
    integer :: h, i, j, k
    fac = [p, p**2/2._wp, p**3/6._wp, al_2pi, al_2pi*p, al_2pi*p**2/2._wp,1._wp]
    lv  = [log_x**3, log_x**2, log_x, 1.0_wp, log_xb, log_xb**2, log_xb**3]
    l0v = [1._wp, ln0]
    aa = 0._wp
    aa(1,:,:,:)=a01
    if (present(a02)) aa(2,:,:,:)=a02
    if (present(a03)) aa(3,:,:,:)=a03
    if (present(a04)) aa(4,:,:,:)=a04
    if (present(a05)) aa(5,:,:,:)=a05
    ! Use Kahan Babushka Klein summation algorithm
    t = 0._wp
    c = 0._wp
    cc = 0._wp
    cs = 0._wp
    ccs = 0._wp
    s = 0._wp
    do h=1, size(aa,1)
      do i=1, size(aa,2)
        do j=1, size(aa,3)
          do k=1, size(aa,4)
            temp = l0v(k)*lv(j)*fac(i)*aa(h,i,j,k)
            t = s + temp
            if (abs(s) .ge. abs(temp)) then
                c = (s - t) + temp
            else
                c = (temp - t) + s
            end if
            s = t
            t = cs + c
            if (abs(cs) .ge. abs(c)) then
                cc = (cs - t) + c
            else
                cc = (c - t) + cs
            end if
            cs = t
            ccs = ccs + cc
          enddo
        enddo
      enddo
    enddo
    res = s + cs + ccs
  end function full_series

  module pure function rec_log_series (log_x, log_xb, ln0, expansion) result (jll_nll)
    real(wp), intent(in) :: log_x, log_xb, ln0
    real(wp), dimension(7,7,2), intent(in) :: expansion
    real(wp), dimension(6) :: jll_nll
    real(wp), dimension(7) :: lv
    real(wp), dimension(2) :: l0v
    real(wp) :: temp, s, t, c , cc, cs, ccs
    integer :: i, j, k
    lv = [log_x**3, log_x**2, log_x, 1.0_wp, log_xb, log_xb**2, log_xb**3]
    l0v = [1._wp, ln0]
    jll_nll = 0._wp
    do i = 1, size(expansion,1) - 1
      ! Use Kahan Babushka Klein summation algorithm
      t = 0._wp
      c = 0._wp
      cc = 0._wp
      cs = 0._wp
      ccs = 0._wp
      s = 0._wp
      do j = 1, size(expansion,2)
        do k = 1, size(expansion,3)
          temp = l0v(k)*lv(j)*expansion(i,j,k)
          t = s + temp
          if (abs(s) .ge. abs(temp)) then
              c = (s - t) + temp
          else
              c = (temp - t) + s
          end if
          s = t
          t = cs + c
          if (abs(cs) .ge. abs(c)) then
              cc = (cs - t) + c
          else
              cc = (c - t) + cs
          end if
          cs = t
          ccs = ccs + cc
        end do
      end do
     jll_nll(i)=s + cs + ccs
    end do
  end function rec_log_series

  module pure function add_logvec (expansion1, expansion2) result (jll_nll)
    real(wp), dimension(7,7,2), intent(in) :: expansion1, expansion2
    real(wp), dimension(7,7,2) :: jll_nll
    integer :: i,j,k
    real(wp) :: temp, s, t, c , cc, cs, ccs
    do i = 1, size(jll_nll,1)
      do j = 1, size(jll_nll,2)
        do k = 1, size(jll_nll,3)
         jll_nll(i,j,k) = expansion1(i,j,k) + expansion2(i,j,k)
        end do
      end do
    end do
  end function add_logvec

  module pure function sub_logvec (expansion1, expansion2) result (jll_nll)
    real(wp), dimension(7,7,2), intent(in) :: expansion1, expansion2
    real(wp), dimension(7,7,2) :: jll_nll
    integer :: i,j,k
    real(wp) :: temp, s, t, c , cc, cs, ccs
    do i = 1, size(jll_nll,1)
      do j = 1, size(jll_nll,2)
        do k = 1, size(jll_nll,3)
         jll_nll(i,j,k) = expansion1(i,j,k) - expansion2(i,j,k)
        end do
      end do
    end do
  end function sub_logvec

  module pure function endpoint_func_S (x, xb, xp, pl, n) result (ints)
    real(wp), intent(in) :: x, xb, xp, n
    real(wp), dimension(12), intent(in) :: pl
    real(wp), dimension(7) :: ints
    real(wp), dimension(7) :: ele, pos
    integer :: i
    ele=endpoint_func_ELE (x, xb, xp, pl, n)
    pos=endpoint_func_POS (x, xb, xp, pl, n)
    do i = 1, size(ints)
      ints(i) = ele(i) + pos(i)
    end do
  end function endpoint_func_S

  module pure function endpoint_func_NS (x, xb, xp, pl, n) result (ints)
    real(wp), intent(in) :: x, xb, xp, n
    real(wp), dimension(12), intent(in) :: pl
    real(wp), dimension(7) :: ints
    real(wp), dimension(7) :: ele, pos
    integer :: i
    ele=endpoint_func_ELE (x, xb, xp, pl, n)
    pos=endpoint_func_POS (x, xb, xp, pl, n)
    do i = 1, size(ints)
      ints(i) = ele(i) - pos(i)
    end do
  end function endpoint_func_NS

  module pure function endpoint_func_GAM (x, xb, xp, pl) result (ints)
    real(wp), intent(in) :: x, xb, xp
    real(wp), dimension(12), intent(in) :: pl
    real(wp), dimension(7) :: ints
    ints=0._wp
    if (xb .le. 1.0e-7_wp) then
      ints(4) = (2._wp*xb*(-pi2 + ln2*(4._wp*pi2 + (9._wp - 10._wp*ln2)*ln2) + &
        3._wp*(8._wp - 11._wp*zeta3)))/3._wp
      ints(5) = (-4._wp*(6._wp - pi2)*xb)/3._wp
      ints(7) = -4._wp*xb
    elseif (x .le. 1.0e-7_wp ) then
      ints(4) = (-8._wp*zeta3)/x
    else
      ints(3) = (8._wp*(-pi2 + 6._wp*pl(6)))/3._wp
      ints(4) = (4._wp*(pl(1)*(pi2 - 3._wp*pl(1)**2) + 6._wp*(2._wp*(pl(5) - pl(9)) -&
        zeta3)))/(3._wp*x) + (2._wp*(2._wp*(7._wp*ln2**3 + pi2*(-3._wp*ln2 - pl(1)) - &
        pl(1)**3 + 6._wp*(-(pl(1)*pl(6)) - pl(1)*pl(8) - 4._wp*(pl(7) + pl(9)))) + &
        39._wp*zeta3))/3._wp
      ints(5) = (-4._wp*pi2)/3._wp + (4._wp*(pi2 - 6._wp*pl(4)))/(3._wp*x)
      ints(6) = 0._wp
      ints(7) = 4._wp - 4._wp/x
    endif
  end function endpoint_func_GAM

  module pure function endpoint_func_POS (x, xb, xp, pl, n) result (ints)
    real(wp), intent(in) :: x, xb, xp, n
    real(wp), dimension(12), intent(in) :: pl
    real(wp), dimension(7) :: ints
    ints = 0._wp
    if (xb .le. 1.0e-7_wp) then
      ints(4) = (xb*(4._wp*(-18._wp*n - pi2) + 15._wp*ln2*(6._wp + 5._wp*ln2)))/6._wp
      ints(5) = -1._wp/3._wp*(pi2*xb)
    elseif (x .le. 1.0e-7_wp) then
      ints(1) = -4._wp/3._wp
      ints(2) = 1._wp
      ints(3) = 2._wp*(n*pi2 - 2._wp*(1._wp - pi)*(1._wp + pi))
      ints(4) = (4._wp*ln2*(2._wp*pi2 + 5._wp*(3._wp - 2._wp*ln2)*ln2) + &
        3._wp*(-2._wp*(8._wp + pi2) + (61._wp + 24._wp*n)*zeta3))/6._wp
    else
      ints(1) = (4._wp*xb)/3._wp - 8._wp/(3._wp*xp)
      ints(2) = -3._wp - 7._wp*x + 4._wp/xp
      ints(3) = (-2._wp*(pi2 + 3._wp*(-pl(1)**2 - 2._wp*(pl(6) + &
        pl(8)))))/(3._wp*xb) + (2._wp*(pi2 + 2._wp*(3._wp*pl(1)**2 + pl(2) + pl(4) + &
        2._wp*(-pl(6) + 3._wp*pl(8)))))/xp + (2._wp*(-(pi2*(2._wp - 3._wp*x)) + &
        3._wp*(n*pi2 - (2._wp - x)*pl(1)**2 - 2._wp*(xp - (2._wp - 3._wp*x)*pl(6) - &
        3._wp*n*pl(2) + (2._wp - x)*pl(8)))))/3._wp
      ints(4) = (4._wp*(-5._wp*ln2**3 + pi2*(2._wp*ln2 - pl(1)) + &
        3._wp*(pl(1)**3 + 2._wp*(-pl(7) + 2._wp*pl(9)))))/(3._wp*xb) - (pi2*(19._wp + &
        15._wp*x + 2._wp*(-2._wp*ln2 - 21._wp*pl(1))) + 6._wp*(pl(1)*(-3._wp*(2._wp - &
        pl(1))*pl(1) + 2._wp*(-5._wp*pl(6) - 4._wp*(pl(2) + pl(4)))) + &
        2._wp*(5._wp*(ln2**3 + pl(1)*pl(8)) + 2._wp*(5._wp*(ln2*pl(10) + pl(11)) + &
        2._wp*(pl(8) - pl(7) + pl(3) + 2._wp*(-pl(5) - 3._wp*pl(9)) + &
        pl(12))))))/(3._wp*xp) - (4._wp*(9._wp + x)*zeta3)/(xb*xp) + (3._wp*(59._wp - &
        19._wp*x)*zeta3 + 2._wp*(pi2*(-8._wp*ln2 - (1._wp - 5._wp*x)*pl(1)) + &
        3._wp*pl(1)**2*(-5._wp*xp - (3._wp + x)*pl(1)) + 2._wp*(-((12._wp - &
        7._wp*pi2)*xb) + 5._wp*(ln2**2*(3._wp*(1._wp + 2._wp*x) + 2._wp*ln2) + &
        3._wp*((-2._wp*x - xb*pl(1))*pl(6) + (2._wp*xp - xb*pl(1))*pl(8))) + &
        6._wp*((-6._wp*n + xb)*pl(3) + (-3._wp*n - 2._wp*xb)*pl(5) - (13._wp - &
        9._wp*x)*pl(9) + xb*pl(12) + 2._wp*(-(xb*pl(1)*pl(2)) - xb*pl(1)*pl(4) - (2._wp - &
        3._wp*x)*pl(7) + 3._wp*n*zeta3)))))/6._wp
      ints(5) = (4._wp*(5._wp*(ln2**2 - pl(1)**2) - 4._wp*pl(6)))/xp + &
        (2._wp*(-pi2 + 12._wp*xb*pl(6)))/3._wp
    endif
  end function endpoint_func_POS

  module pure function endpoint_func_ELE (x, xb, xp, pl, n) result (ints)
    real(wp), intent(in) :: x, xb, xp, n
    real(wp), dimension(12), intent(in) :: pl
    real(wp), dimension(7) :: ints
    ints=0._wp
    if (xb .le. 1.0e-7_wp) then
      ints(4) = ((-6._wp*(5._wp + 6._wp*n) - pi2)*xb)/3._wp
      ints(5) = 4._wp*xb
    elseif (x .le. 1.0e-7_wp) then
      ints(3) = (-2._wp*(1._wp - 3._wp*n)*pi2)/3._wp
      ints(4) = 2._wp*(-pi2 - 2._wp*(1._wp - 3._wp*n*zeta3))
    else
      ints(2) = -2._wp*x
      ints(3) = -1._wp/3._wp*(21._wp + 4._wp*pi2)/xb + (pi2*xp + 3._wp*(7._wp + &
        x + 2._wp*(n*pi2 + (6._wp*n + xp)*pl(2) + xp*pl(4))))/3._wp
      ints(4) = (-(pi2*(7._wp - 10._wp*x)) + 3._wp*(5._wp*xb - &
        6._wp*pl(2)))/(3._wp*xb) + (12._wp*(-((6._wp*n + xp)*pl(3)) - (3._wp*n + &
        xp)*pl(5) + 6._wp*n*zeta3) + xp*(pi2 - 3._wp*(9._wp - 4._wp*(pl(2) + &
        zeta3))))/3._wp
      ints(5) = -3._wp*(1._wp - 3._wp*x) + 4._wp*xp*pl(4)
    endif
  end function endpoint_func_ELE

  subroutine photon_matching (log_xb, x0, x1, p, bar, asym)
    real(wp), intent(in) :: log_xb, x0, x1, p
    real(wp), intent(inout), dimension(7,7,2) :: bar
    real(wp), intent(inout), dimension(7,7,2) :: asym
    real(wp) :: xm, logxb10, p_match
    logxb10 = - log_xb / ln10
    if (logxb10 < x0) then
       bar = bar
       asym = 0._wp
    else if (logxb10 > x1) then
       bar = 0._wp
       asym = asym
    else
       xm = (x1 - x0) / (logxb10 - x0)
       p_match = (1._wp + (xm - 1._wp)**p)**(-1)
       bar = (1._wp - p_match) * bar
       asym = p_match * asym
    end if
  end subroutine photon_matching

  module function elec_pdf (epdf, flv, x, xb, scale) result (res)
    type(qed_pdf_t), intent(in) :: epdf
    real(wp) :: res
    integer, intent(in) :: flv
    real(wp), intent(in) :: x, xb
    real(wp), intent(in) :: scale
    integer :: nf, nlep
    real(wp), parameter :: &
      x0gam = 2.0_wp, x1gam = 6.0_wp, pgam = 2.0_wp
    real(wp), parameter :: xmin=0._wp, xmax=1._wp
    real(wp) :: xp, log_x, log_xb
    real(wp) :: ln0, eta0, alpha, al0_2pi, al_2pi, p, n, run
    real(wp), dimension(7,7,2) :: &
      asym=0._wp, asymb=0._wp, hat=0._wp, bar=0._wp, num=0.0_wp
    real(wp), dimension(12) :: pl
    logical :: running
    logical, dimension(6) :: order
    ! Cut for x->0
    if (x .le. 1.0e-45_wp) then
      res = 0._wp
      return
    end if
    ! Cut Positron part for x->1
    if (xb .le. 1.0e-6_wp .and. flv .eq. EPDF_POS) then
      res = 0._wp
      return
    end if
    ! Cut Photon part for x->1
    if (xb .le. 1.0e-8_wp .and. flv .eq. EPDF_G) then
      res = 0._wp
      return
    end if
    ! Cut Electron part for x->1
    if (xb .le. 1.0e-15_wp .and. flv .eq. EPDF_ELE) then
      res = 0._wp
      return
    end if
    ! Calculate computing intensive polylogs only if necessary
    pl = 0._wp
    if (xb .ge. 1.0e-7_wp .and. x .ge. 1.0e-7) then
      pl(1) = log(1._wp+x)
      pl(2) = polylog(2,x)
      pl(3) = polylog(3,x)
      pl(4) = polylog(2,xb)
      pl(5) = polylog(3,xb)
      pl(6) = polylog(2,-x)
      pl(7) = polylog(3,-x)
      pl(8) = polylog(2,1._wp/(1._wp+x))
      pl(9) = polylog(3,1._wp/(1._wp+x))
      pl(10) = polylog(2,(1._wp+x)/2._wp)
      pl(11) = polylog(3,(1._wp+x)/2._wp)
      pl(12) = polylog(3,xb*(1._wp+x))
    endif
    xp = 1._wp + x
    log_x = log_prec(x,xb)
    log_xb = log_prec(xb,x)
    call set_qed_pdf_parameters (epdf, scale, alpha, running, &
      nlep, nf, ln0, eta0, p, al0_2pi, al_2pi, n, run, order)
    select case (flv)
    case (EPDF_S,EPDF_NS,EPDF_ELE, EPDF_POS)
       bar   = recbar (flv, x, xb, n, run, order)
       hat   = rechat (flv, x, xb, xp, pl, n, run, order)
       asymb = - bar_asym (flv, xb, n, run, order)
       asym  = recasym (flv, x, xb, log_xb, nlep, nf, &
         n, p, al0_2pi, al_2pi, order, running)
       num = recnum (flv, x, xb, xp, pl, n)
    case (EPDF_G)
       bar   = recbar (flv, x, xb, n, run, order)
       hat   = rechat (flv, x, xb, xp, pl, n, run, order)
       asymb = 0._wp
       asym  = recasym (flv, x, xb, log_xb, nlep, nf, &
         n, p, al0_2pi, al_2pi, order, running)
       num = recnum (flv, x, xb, xp, pl, n)
       call photon_matching (log_xb, x0gam, x1gam, pgam, bar, asym)
    case default
       call msg_fatal ("elec_pdf: wrong lepton flavor.")
    end select
    res = full_series (log_x, log_xb, ln0, p, al_2pi, bar, hat, num, asymb, asym)
  end function elec_pdf


end submodule electron_pdfs_s
